
// Arduino interface code for ARDOP running on a Teensy 3.6

#include "TeensyConfig.h"
#include "TeensyCommon.h"

#ifndef ARDOP
#error("ARDOP not defined in TeensyConfig.h");
#endif

#ifndef PACKET
#error("PACKET not defined in TeensyConfig.h");
#endif


#define CPU_RESTART_ADDR (uint32_t *)0xE000ED0C
#define CPU_RESTART_VAL 0x5FA0004
#define CPU_RESTART (*CPU_RESTART_ADDR = CPU_RESTART_VAL);

#include <EEPROM.h>

unsigned char RXBUFFER[500];	// Async RX Buffer. Enough for Stuffed Host Mode Frame

extern volatile int RXBPtr;
volatile int flag = 0;
volatile int flag2 = 0;
volatile int lastperiod = 0, lastint = 0;

int SerialHost = TRUE;				// Will eventually allow switching from Serial to I2c without reconfig
int ActivePort = 0;						// Serial port in use
int SerialWatchDog = 0;

extern int VRef;
extern int TXLevel;
extern int autoRXLevel;

void CommonSetup();
void setupPDB(int SampleRate);
void setupDAC();
void setupADC(int Pin);
extern "C" void StartDAC();
extern "C" void StopDAC();
void StartADC();
void setupOLED();
void setupKMR_18();

void setupWDTTFT();

// Arduino is a c++ environment so functions in ardop must be defined as "C"

extern "C"
{
  void WriteDebugLog(int Level, const char * format, ...);
  void InitSound();
  void HostInit();
  void CheckTimers();
  void HostPoll();
  void MainPoll();
  void InitDMA();
  void PlatformSleep();
  void AdjustTXLevel(int Level);
  void KISSInit();

  void ProcessSCSPacket(unsigned char * rxbuffer, int Length);

#include "../../ARDOPC/ARDOPC.h"

  extern unsigned int tmrPollOBQueue;

  extern int UseKISS;

#define Now getTicks()

#define MEM_LEN 512
  extern uint8_t databuf[MEM_LEN];
  extern volatile int i2cputptr, i2cgetptr;

  void HostPoll()
  {
#ifdef I2CHOST

    while (i2cgetptr != i2cputptr)
    {
      unsigned char c;
      c = databuf[i2cgetptr++];
      i2cgetptr &= 0x1ff;				// 512 cyclic

      RXBUFFER[RXBPtr++] = c;

      if (i2cgetptr == i2cputptr || RXBPtr > 498)
      {
        ProcessSCSPacket(RXBUFFER, RXBPtr);
        return;
      }
    }
#else

#ifdef i2cSlaveSupport
    i2cloop();					// I2C but not for Host
#endif

    // To simplify switching USB / Serial / Bluetooth / ESP we poll all defined ports,
    // and switch to using whichever provides input

    // Both boards could use USB(Serial) or Serial1
    // PIBoard has ESP on Serial3
    // WDT Board has BT on Serial5

    int Avail = HOSTPORT.available();
    int Count;

    if (Avail)
    {
      if (Avail > (499 - RXBPtr))
        Avail = 499 - RXBPtr;

      Count = HOSTPORT.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
      RXBPtr += Count;
      if (ActivePort != 1)
      {
        ActivePort = 1;
        WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
      }

      ProcessSCSPacket(RXBUFFER, RXBPtr);
    }
#ifdef HOSTPORT2
    else
    {
      Avail = HOSTPORT2.available();
      if (Avail)
      {
        if (Avail > (499 - RXBPtr))
          Avail = 499 - RXBPtr;

        Count = HOSTPORT2.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
        RXBPtr += Count;
        if (ActivePort != 2)
        {
          ActivePort = 2;
          WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
        }
        ProcessSCSPacket(RXBUFFER, RXBPtr);
      }
    }
#endif
#ifdef HOSTPORT3
    else
    {
      Avail = HOSTPORT3.available();
      if (Avail)
      {
        if (Avail > (499 - RXBPtr))
          Avail = 499 - RXBPtr;

        Count = HOSTPORT3.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
        RXBPtr += Count;
        if (ActivePort != 3)
        {
          ActivePort = 3;
          WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
        }
        ProcessSCSPacket(RXBUFFER, RXBPtr);
      }
    }
#endif
#endif		// I2CHOST
  }
}

void setup()
{
  uint32_t i, sum = 0;

#ifdef I2CHOST
  SerialHost = FALSE;
#endif

  TXLevel = 30;
  RXLevel = 0;
  autoRXLevel = 1500;

  CommonSetup();

  WriteDebugLog(LOGALERT, "ARDOPC Version %s CPU %d Bus %d FreeRAM %d", ProductVersion, F_CPU, F_BUS, FreeRam());
  blnTimeoutTriggered = FALSE;
  SetARDOPProtocolState(DISC);

  InitSound();
  HostInit();
  tmrPollOBQueue = Now + 10000;
  ProtocolMode = ARQ;

  if (UseKISS)
    KISSInit();

  // Configure the ADC and run at least one software-triggered
  // conversion.  This completes the self calibration stuff and
  // leaves the ADC in a state that's mostly ready to use

  analogReadRes(16);
  //analogReference(INTERNAL); // range 0 to 1.2 volts
  analogReference(DEFAULT); // range 0 to 3.3 volts
  //analogReadAveraging(8);
  // Actually, do many normal reads, to start with a nice DC level

  for (i = 0; i < 1024; i++)
  {
    sum += analogRead(16);
  }

  WriteDebugLog(LOGDEBUG, "DAC Baseline %d", sum / 1024);

  setupPDB(12000);			// 12K sample rate
  setupDAC();
  setupADC(16);
  StartADC();

#ifdef PIBOARD

  // PI Baord can read Analog out which allows u to Read Vref

  SetPot(1, 256);				// TX Level Gain = 1

  delay(100);

  analogRead(17);

  for (i = 0; i < 100; i++)
  {
    VRef += analogRead(17);
  }
  VRef /= 100;
  analogRead(16);		// Set ADC back to A0

#endif


  WriteDebugLog(0, "VREF %d offset %d", VRef, VRef - 32768);

  AdjustTXLevel(TXLevel);

  print_mac();

#ifdef PLOTCONSTELLATION
#ifdef OLED
  setupOLED();
#endif
#ifdef WDTTFT
  setupWDTTFT();
#endif
#ifdef KMR_18
  setupKMR_18();
#endif

#endif
}

int lastticks = Now;

void loop()
{
  PollReceivedSamples();
  CheckTimers();
  HostPoll();
  MainPoll();

  if (Now - lastticks > 999)
  {
    // Debug 1 sec tick
  }
  PlatformSleep();
  //RadioPoll();
  Sleep(1);

#ifdef I2CHOST
  SerialWatchDog++;		// Reset if nothing for 60 secs
#endif

  if (blnClosing || SerialWatchDog > 60000)
  {
    CPU_RESTART // reset processor
  }
}

extern "C"
{
  int OKtoAdjustLevel()
  {
    // Only auto adjust level when disconnected.
    // Level is set at end of each received packet when connected

    return (ProtocolState == DISC);
  }

  void StartCapture()
  {
    Capturing = TRUE;
    DiscardOldSamples();
    ClearAllMixedSamples();
    State = SearchingForLeader;
  }
  void StopCapture()
  {
    Capturing = FALSE;
  }
  void TurnroundLink()
  {
    if (blnEnbARQRpt > 0 || blnDISCRepeating)	// Start Repeat Timer if frame should be repeated
      dttNextPlay = Now + intFrameRepeatInterval;
  }
}



// function from the sdFat library (SdFatUtil.cpp)
// licensed under GPL v3
// Full credit goes to William Greiman.

extern "C" char* sbrk(int incr);

extern char *__brkval;
extern char __bss_end;

int FreeRam()
{
  char top;

  return __brkval ? &top - __brkval : &top - &__bss_end;
}

// Function to get mac/serial number

uint8_t mac[8];

// http://forum.pjrc.com/threads/91-teensy-3-MAC-address

void readserialno(uint8_t *mac)
{
  noInterrupts();

  // With 3.6 have to read a block of 8 bytes at zero
  // With 3.1 two blocks of 4 at e and f

  FTFL_FCCOB0 = 0x41;             // Selects the READONCE command
#if defined(__MK64FX512__) || defined(__MK66FX1M0__)
  FTFL_FCCOB1 = 0;             // read the given word of read once area
#else
  FTFL_FCCOB1 = 0x0F;          // read the given word of read once area
#endif
  FTFL_FCCOB2 = 0;
  FTFL_FCCOB3 = 0;
  // launch command and wait until complete
  FTFL_FSTAT = FTFL_FSTAT_CCIF;
  while (!(FTFL_FSTAT & FTFL_FSTAT_CCIF));

#if defined(__MK64FX512__) || defined(__MK66FX1M0__)

  // Top 3 bytes (Manufacture ID in bits 8-31

  *(mac++) = FTFL_FCCOB5;
  *(mac++) = FTFL_FCCOB6;
  *(mac++) = FTFL_FCCOB7;

  // Serial No in bits 8-31

  *(mac++) = FTFL_FCCOB9;
  *(mac++) = FTFL_FCCOBA;
  *(mac++) = FTFL_FCCOBB;
#else

  // Could read Manufacture ID from page E but no point as fixed (04:E9:E5)

  *(mac++) = 0x04;
  *(mac++) = 0xE9;
  *(mac++) = 0xE5;
  *(mac++) = FTFL_FCCOB5;
  *(mac++) = FTFL_FCCOB6;
  *(mac++) = FTFL_FCCOB7;


#endif

  interrupts();
}

void print_mac()
{
  readserialno(mac);
  WriteDebugLog(6, "Hardware Serial No %02X:%02X:%02X:%02X:%02X:%02X",
                mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
}

// Stuff added from SM_Teensy

#define DCDLED LED0
#define PKTLED LED3		// flash when packet received

extern "C" int displayDCD(int state)
{
  SetLED(DCDLED, state);
  return true;
}

// Seem to be needed to satify linker

extern "C" int _gettimeofday()
{
  return 0;
}


extern "C" {
 
 // Dummies for ARDOP GUI

  void DrawTXFrame(const char * Frame)
  {}

  void DrawRXFrame(int State, const char * Frame)
  {}

  int SendtoGUI(char Type, unsigned char * Msg, int Len)
  {
    return 0;
  }
}

// Packet Code

// TNC Params in EEPROM - compatible with pitnc_get/set

#define  TXDELAY   1
#define PERSIST   2
#define SLOTTIME  3
#define TXTAIL    4
#define FULLDUP   5
#define KISSCHANNEL 6
#define I2CADDRESS  7


extern int Baud;    // Modem Speed 300 1200 2400 AFSK 3600 4800 9600 19200 FSK RUH
extern int AFSK;    // Modem Mode
extern int FSK;
extern int PSK;
extern int centreFreq;  // Used for AFSK
extern int samplerate;
extern char VersionString[];
extern int VersionNo;
extern int KISSCHECKSUM;

extern int VRef;

extern int TXLevel;       // 300 mV p-p Used on Teensy
extern int RXLevel;       // Configured Level - zero means auto tune
extern int autoRXLevel;   // calculated level
extern int ARDOPMode;

// Default KISS params. Will be overridden from EEPROM if set

unsigned int txdelay = 300;
unsigned int ppersist = 64;
unsigned int slottime = 100;
unsigned int fullduplex = 0;
unsigned int txtail = 0;

void i2csetup();
void i2cloop();

extern "C"
{
  void SoundModemInit();
  void ProcessKISSPacket(unsigned char * rxbuffer, int Length);
  void mainLoop();
  void SetDefaultKISSParams(unsigned int txdelay, unsigned int ppersist,
                                     unsigned int slottime, unsigned int fullduplex, unsigned int txtail);
}


void pktsetup()
{
  uint32_t i, sum = 0;

  ARDOPMode = 0;
  Baud = 9600;    // FSK G3RUH
  AFSK = FALSE;
  FSK = TRUE;
  PSK = FALSE;

  // if EEPROM KISS Params are set use them to overide defaults
  // unset EEPROM defaults to 255

  int value = EEPROM.read(TXDELAY);

  if (value != 255)
    txdelay = value * 10;

  value = EEPROM.read(PERSIST);

  if (value != 255)
    ppersist = value;

  value = EEPROM.read(SLOTTIME);

  if (value != 255)
    slottime = value * 10;

  value = EEPROM.read(TXTAIL);

  if (value != 255)
    txtail = value;

  value = EEPROM.read(FULLDUP);

  if (value == 1)
    fullduplex = 1;

  SetDefaultKISSParams(txdelay,  ppersist, slottime, fullduplex, txtail);

  // if EEPROM Reg 8 is set to a valid speed set TNC mode
  // otherwise use compiled in defaults

  value = EEPROM.read(8);

  if (value == 3)
  {
    Baud = 300;
    AFSK = TRUE;
    FSK = FALSE;
  }
  else if (value == 12)
  {
    Baud = 1200;
    AFSK = TRUE;
    FSK = FALSE;
  }
  else if (value == 24)
  {
    Baud = 2400;
    AFSK = TRUE;
    FSK = FALSE;
  }
  else if (value == 36)
  {
    Baud = 3600;  // FSK G3RUH
    AFSK = FALSE;
    FSK = TRUE;
  }
  else if (value == 48)
  {
    Baud = 4800;  // FSK G3RUH
    AFSK = FALSE;
    FSK = TRUE;
  }
  else if (value == 72)
  {
    Baud = 7200;  // FSK G3RUH
    AFSK = FALSE;
    FSK = TRUE;
  }

  else if (value == 96)
  {
    Baud = 9600;  // FSK G3RUH
    AFSK = FALSE;
    FSK = TRUE;
  }
  else if (value == 191)    // Cant use 182 in SetParams
  {
    Baud = 19200; // FSK G3RUH
    AFSK = FALSE;
    FSK = TRUE;
  }

  // if EEPROM Reg 13 is set to a reasonable value set TNC centre freq
  // otherwise use compiled in defaults

  value = EEPROM.read(13);

  if (value > 90 && value < 250)
    centreFreq = value * 10;

  RXLevel = EEPROM.read(9);
  TXLevel = EEPROM.read(10);
  autoRXLevel = 1500;

  CommonSetup();
  print_mac();

#ifdef TFT
  TFTprintf("Packet TNC based on Soundmodem by Thomas Sailer");
#endif
  MONprintf(VersionString);

  SaveEEPROM(0, VersionNo);

  WriteDebugLog(7, "CPU %d Bus %d FreeRAM %d", F_CPU, F_BUS, FreeRam());

  for (i = 0; i < 16; i++)
  {
    int value = EEPROM.read(i);
    WriteDebugLog(7, "%d\t%d", i, value);
  }

  if (PSK)
  {
#ifdef TFT
    TFTprintf("PSK Mode");
#endif
    MONprintf("PSK Mode");
  }
  else if (FSK)
  {
#ifdef TFT
    TFTprintf("9600 FSK Mode");
#endif
    MONprintf("9600 FSK Mode");
  }
  else
  {
#ifdef TFT
    TFTprintf("AFSK Mode %d Baud Centre Freq %d", Baud, centreFreq);
#endif
    MONprintf("AFSK Mode %d Baud Centre Freq %d", Baud, centreFreq);
  }

  if (RCM_SRS0 & 0X20)    // Watchdog Reset
    WriteDebugLog(LOGCRIT, "\n**** Reset by Watchdog ++++");

  SoundModemInit();

  // Configure the ADC and run at least one software-triggered
  // conversion.  This completes the self calibration stuff and
  // leaves the ADC in a state that's mostly ready to use

  analogReadRes(16);
  //analogReference(INTERNAL); // range 0 to 1.2 volts
  analogReference(DEFAULT); // range 0 to 3.3 volts
  //analogReadAveraging(8);
  // Actually, do many normal reads, to start with a nice DC level

  for (i = 0; i < 1024; i++)
  {
    sum += analogRead(16);
  }

  WriteDebugLog(LOGDEBUG, "DAC Baseline %d", sum / 1024);

#ifdef PIPOARD

  // Can read DAC output to determine Vref

  SetPot(1, 256);       // TX Level Gain = 1

  delay(100);

  analogRead(17);

  for (i = 0; i < 100; i++)
  {
    VRef += analogRead(17);
  }
  VRef /= 100;
  MONprintf("VREF %d offset %d\r\n", VRef, VRef - 32768);

  analogRead(16);   // Set ADC back to A0
#endif

  SetPot(1, TXLevel);
  WriteDebugLog(LOGDEBUG, "TXLevel %d = %d mV", TXLevel, (TXLevel * 3000) / 256);

  if (RXLevel)
    SetPot(0, RXLevel);
  else
    SetPot(0, 256);         // Start at min gain

  setupPDB(samplerate);
  setupDAC();
  setupADC(16);
  StartADC();

  WriteDebugLog(7, "CPU %d Bus %d FreeRAM %d", F_CPU, F_BUS, FreeRam());

  //  if (PSK)
  //    RunPSKReceive();    // This does not return!!!!
}


void pktloop()
{
  mainLoop();
  PlatformSleep();
  Sleep(1);
  HostPoll();
#if defined I2CKISS || defined I2CMONITOR
  i2cloop();
#endif
}
